import json


def get_params_for_repr(params):
    param_list = []
    for name, param in sorted(params.items()):
        param_str = json.dumps(param)
        if param_str == "true":
            param_str = "True"
        if param_str == "false":
            param_str = "False"
        param_list.append(f"{name}={param_str}")
    return "\n        " + ",\n        ".join(param_list)



import re
from typing import Any
from typing import Dict

from email_validator import EmailNotValidError
from email_validator import validate_email
from simpleeval import EvalWithCompoundTypes as Evaluator
from simpleeval import InvalidExpression

from .questions import Validator


class ValidationError(Exception):
    """
    Validation error exception, for use by Form.update_object.
    """


def text_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Validate length of a text value, and whether digits are allowed.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    max_length = int(validator.max_length)
    min_length = int(validator.min_length)
    allow_digits = validator.allow_digits
    result = True
    value = str(value)
    length = len(value)
    if length < min_length or max_length > 0 and length > max_length:
        result = False
    if not allow_digits:
        for number in range(10):
            if str(number) in value:
                result = False
    return result


def numeric_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Validate if number value is within set limits.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    max_value = float(validator.max_value)
    min_value = float(validator.min_value)
    result = True
    value = float(value)
    if value < min_value or max_value > 0 and value > max_value:
        result = False
    return result


def email_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Validate if value is a valid email address.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    result = True
    try:
        validate_email(value)
    except EmailNotValidError:
        result = False
    return result


def regex_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Validate if value matches regular expression.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    regex = validator.regex
    regex = re.compile(regex)
    return regex.match(value) is not None


def expression_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Validate if expression associated with value is true.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    expression = validator.expression
    result = True
    if expression:
        expression = expression.replace("{", "").replace("}", "")
        expression = expression.replace(" empty", " in [[], {}]")
        expression = expression.replace(" notempty", " not in [[], {}]")
        expression = expression.replace(" anyof ", " in ")
        try:
            evaluator = Evaluator(names=form_data)
            result = evaluator.eval(expression) is True
        except (ValueError, TypeError, SyntaxError, KeyError, InvalidExpression):
            result = False
    return result


VALIDATORS = {
    "text": text_validator,
    "numeric": numeric_validator,
    "email": email_validator,
    "regex": regex_validator,
    "expression": expression_validator,
}


def call_validator(validator: Validator, value: Any, form_data: Dict[str, Any]):
    """Call correct validation method depending on validator type.

    :param validator:
        The validator instance for the current question.
    :param value:
        The value to be validated.
    :param form_data:
        The dictionary containing from data entered for current form.

    :Returns:
        If validation passes, :data:`True`, else :data:`False`.
    """
    validator_method = VALIDATORS[validator.kind]
    return validator_method(validator, value, form_data)



#!/usr/bin/env python

"""Tests for `templates` package."""

from questions import templates
from questions.settings import BOOTSTRAP_URL
from questions.settings import SURVEY_JS_CDN


def test_get_platform_js_resources():
    js = templates.get_platform_js_resources("jquery", SURVEY_JS_CDN)
    assert f"{SURVEY_JS_CDN}/survey-jquery/survey.jquery.min.js" in js
    js = templates.get_platform_js_resources("jquery", "http://testing")
    assert "http://testing/survey-jquery/survey.jquery.min.js" in js


def test_get_theme_css_resources():
    css = templates.get_theme_css_resources("default", SURVEY_JS_CDN)
    assert f"{SURVEY_JS_CDN}/survey-core/survey.min.css" in css
    css = templates.get_theme_css_resources("bootstrap", SURVEY_JS_CDN)
    assert BOOTSTRAP_URL in css
    css = templates.get_theme_css_resources("default", "http://testing")
    assert "http://testing/survey-core/survey.min.css" in css
    css = templates.get_theme_css_resources("modern", "http://testing")
    assert "http://testing/survey-core/modern.min.css" in css
    css = templates.get_theme_css_resources("bootstrap", "http://testing")
    assert "http://testing/bootstrap.min.css" in css


def test_get_survey_js():
    survey_js = templates.get_survey_js(
        form_json="FORM_JSON",
        form_data="FORM_DATA",
        html_id="id",
        action="http://testing",
        theme="defaultV2",
        platform="jquery",
    )
    assert "FORM_JSON" in survey_js
    assert "FORM_DATA" in survey_js
    assert "http://testing" in survey_js
    assert "defaultV2" in survey_js


def test_get_survey_js_form_data_is_none():
    survey_js = templates.get_survey_js(
        form_json="FORM_JSON",
        form_data=None,
        html_id="id",
        action="http://testing",
        theme="defaultV2",
        platform="jquery",
    )
    assert "FORM_JSON" in survey_js
    assert "http://testing" in survey_js
    assert "defaultV2" in survey_js


def test_get_form_page():
    html = templates.get_form_page(
        title="Title",
        html_id="id",
        platform="jquery",
        survey_js="GENERATED_JS",
        js_resources=["file.js"],
        css_resources=["file.css"],
    )
    assert "Title" in html
    assert "GENERATED_JS" in html
    assert "file.js" in html
    assert "file.css" in html


def test_get_form_page_no_resources():
    html = templates.get_form_page(
        title="Title",
        html_id="id",
        platform="jquery",
        survey_js="GENERATED_JS",
        js_resources=None,
        css_resources=None,
    )
    assert "Title" in html
    assert "GENERATED_JS" in html



"""
Console scripts for questions.

 - list_resources PLATFORM THEME [--include-widgets]
     Lists all CSS and JS resources needed to run SurveyJS

 - download_surveyjs PATH PLATFORM THEME
     Downloads all CSS and JS resources needed to run SurveyJS
"""
import click
import requests

from .form import Form
from .form import FormPage
from .form import FormPanel
from .questions import Question
from .questions import QUESTION_TYPES
from .settings import SURVEY_JS_CDN
from .templates import get_platform_js_resources
from .templates import get_theme_css_resources


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.argument("platform")
@click.argument("theme")
def download_surveyjs(path, platform, theme):
    """
    Download SurveyJS files to specified location.

    Supplied path must exist.

    Platforms are: angular, jquery, knockout, react, vue.

    Themes are: default, defaultV2, bootstrap, darkblue, darkrose, modern, orange, stone
                winter, winterstone.
    """
    click.echo()
    js = get_platform_js_resources(platform=platform, resource_url=SURVEY_JS_CDN)
    css = get_theme_css_resources(theme=theme, resource_url=SURVEY_JS_CDN)
    widgets_js = set()
    widgets_css = set()
    for question in QUESTION_TYPES:
        extra_js = question.__fields__.get("extra_js").get_default()
        widgets_js.update(extra_js)
        extra_css = question.__fields__.get("extra_css").get_default()
        widgets_css.update(extra_css)
    for url in js + css + list(widgets_js) + list(widgets_css):
        resource = requests.get(url, allow_redirects=True)
        filename = path + "/" + resource.url.split("/")[-1]
        with open(filename, "wb") as downloaded:
            downloaded.write(resource.content)
            click.echo(f"Downloaded {filename}")
    click.echo()


@click.command()
@click.argument("platform")
@click.argument("theme")
@click.option(
    "--include-widgets", default=False, help="include widget resources", is_flag=True
)
def list_resources(platform, theme, include_widgets):
    """
    List required resources for a platform and theme.

    Platforms are: angular, jquery, knockout, react, vue.

    Themes are: default, defaultV2, bootstrap, darkblue, darkrose, modern, orange, stone
                winter, winterstone.
    """
    click.echo()
    js = get_platform_js_resources(platform=platform, resource_url=SURVEY_JS_CDN)
    css = get_theme_css_resources(theme=theme, resource_url=SURVEY_JS_CDN)
    click.echo("Required Javascript resources:")
    for url in js:
        click.echo(f"    {url}")
    if include_widgets:
        click.echo()
        click.echo("Widget specific Javascript resources:")
        for question in QUESTION_TYPES:
            extra_js = question.__fields__.get("extra_js").get_default()
            if extra_js:
                click.echo()
                click.echo(f"{question.__name__}:")
                for url in extra_js:
                    click.echo(f"    {url}")
        click.echo()
    click.echo()
    click.echo("Required CSS resources:")
    for url in css:
        click.echo(f"    {url}")
    if include_widgets:
        click.echo()
        click.echo("Widget specific CSS resources:")
        for question in QUESTION_TYPES:
            extra_css = question.__fields__.get("extra_css").get_default()
            if extra_css:
                click.echo()
                click.echo(f"{question.__name__}:")
                for url in extra_css:
                    click.echo(f"    {url}")
    click.echo()


@click.command()
@click.argument("name")
@click.argument("json_file", type=click.File("r"))
def generate_code(name, json_file):
    """
    Generate Questions form code from SurveyJS JSON file.
    """
    imports = ["from questions import Form"]
    code = []
    NewForm = Form.from_json(json_file.read(), name)
    current = [NewForm]
    while current != []:
        fragment = []
        next_batch = []
        for member in current:
            fragment.append(f"\n\nclass {member.__name__}(Form):\n")
            for element_name, element in member.__dict__.items():
                if isinstance(element, (FormPage, FormPanel)):
                    next_batch.append(element.form.__class__)
                if isinstance(element, (FormPage, FormPanel, Question)):
                    statement = f"from questions import {element.__class__.__name__}"
                    if statement not in imports:
                        imports.append(statement)
                if element_name.startswith("_"):
                    continue
                element_repr = repr(element)
                element_repr = element_repr.replace(": false,", ": False,")
                element_repr = element_repr.replace(": true,", ": True,")
                element_repr = element_repr.replace(": false}", ": False}")
                element_repr = element_repr.replace(": true}", ": True}")
                fragment.append(f"    {element_name} = {element_repr}\n")
        code.append(fragment)
        current = next_batch
    click.echo()
    for statement in imports:
        click.echo(statement)
    code.reverse()
    for fragment in code:
        click.echo("".join(fragment))



"""Top-level package for Questions."""

__author__ = """Carlos de la Guardia"""
__email__ = "cguardia@yahoo.com"
__version__ = "0.1.0"


from .form import Form
from .form import FormPage
from .form import FormPanel
from .questions import BarRatingQuestion
from .questions import BooleanQuestion
from .questions import BootstrapDatePickerQuestion
from .questions import BootstrapSliderQuestion
from .questions import CheckboxQuestion
from .questions import CKEditorQuestion
from .questions import CommentQuestion
from .questions import DropdownQuestion
from .questions import EmailValidator
from .questions import EmotionsRatingQuestion
from .questions import ExpressionBlock
from .questions import ExpressionValidator
from .questions import FileQuestion
from .questions import HtmlBlock
from .questions import ImageBlock
from .questions import ImagePickerQuestion
from .questions import JQueryUIDatePickerQuestion
from .questions import MatrixDropdownQuestion
from .questions import MatrixDynamicQuestion
from .questions import MatrixQuestion
from .questions import MicrophoneQuestion
from .questions import MultipleTextQuestion
from .questions import NoUISliderQuestion
from .questions import NumericValidator
from .questions import RadioGroupQuestion
from .questions import RatingQuestion
from .questions import RegexValidator
from .questions import Select2Question
from .questions import SignaturePadQuestion
from .questions import SortableJSQuestion
from .questions import TagBoxQuestion
from .questions import TextQuestion
from .questions import TextValidator
from .settings import SURVEY_JS_CDN
from .validators import ValidationError


__all__ = [
    "SURVEY_JS_CDN",
    "BarRatingQuestion",
    "BooleanQuestion",
    "BootstrapDatePickerQuestion",
    "BootstrapSliderQuestion",
    "CheckboxQuestion",
    "CKEditorQuestion",
    "CommentQuestion",
    "DropdownQuestion",
    "EmailValidator",
    "EmotionsRatingQuestion",
    "ExpressionBlock",
    "ExpressionValidator",
    "FileQuestion",
    "Form",
    "FormPage",
    "FormPanel",
    "HtmlBlock",
    "ImageBlock",
    "ImagePickerQuestion",
    "JQueryUIDatePickerQuestion",
    "MatrixDropdownQuestion",
    "MatrixDynamicQuestion",
    "MatrixQuestion",
    "MicrophoneQuestion",
    "MultipleTextQuestion",
    "NoUISliderQuestion",
    "NumericValidator",
    "RadioGroupQuestion",
    "RatingQuestion",
    "RegexValidator",
    "Select2Question",
    "SignaturePadQuestion",
    "SortableJSQuestion",
    "TagBoxQuestion",
    "TextQuestion",
    "TextValidator",
    "ValidationError",
]



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "email_validator",
    "Jinja2",
    "pydantic",
    "requests",
    "simpleeval",
]

setup(
    author="Carlos de la Guardia",
    author_email="cguardia@yahoo.com",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description="Questions is a form library that uses the power of SurveyJS for the UI.",
    entry_points={
        "console_scripts": [
            "download_surveyjs=questions.cli:download_surveyjs",
            "list_resources=questions.cli:list_resources",
            "generate_code=questions.cli:generate_code",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="questions",
    name="questions",
    packages=find_packages(include=["questions", "questions.*"]),
    test_suite="tests",
    url="https://github.com/cguardia/questions",
    version="0.8.0",
    zip_safe=False,
)



#!/usr/bin/env python

"""Tests for `form` package."""
import pytest

from questions import form
from questions import questions
from questions import TextValidator
from questions import ValidationError
from questions.settings import SURVEY_JS_CDN


def test_initialize():
    test_form = form.Form(name="testing", param1=0)
    assert test_form.name == "testing"
    assert "param1" in test_form.params


def test_initialize_no_name():
    test_form = form.Form()
    assert test_form.name == "Form"


def test_call():
    test_form = form.Form()
    assert test_form() == test_form.render_html()


def test_construct_survey():
    class TestForm(form.Form):
        text1 = questions.TextQuestion()

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements


def test_construct_survey_with_extra_js():
    class TestForm(form.Form):
        text1 = questions.Select2Question()

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "select2.min.js" in test_form.extra_js[0]
    assert "cdnjs" in test_form.extra_js[0]


def test_construct_survey_with_extra_css():
    class TestForm(form.Form):
        text1 = questions.TagBoxQuestion()

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "select2.min.css" in test_form.extra_css[0]
    assert "cdnjs" in test_form.extra_css[0]


def test_construct_survey_with_extra_js_and_resource_url():
    class TestForm(form.Form):
        text1 = questions.Select2Question()

    test_form = TestForm(name="testing", resource_url="static")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "static/select2.min.js" in test_form.extra_js[0]


def test_construct_survey_with_extra_css_and_resource_url():
    class TestForm(form.Form):
        text1 = questions.TagBoxQuestion()

    test_form = TestForm(name="testing", resource_url="static")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "static/select2.min.css" in test_form.extra_css[0]


def test_construct_survey_with_extra_js_and_set_resource_url():
    class TestForm(form.Form):
        text1 = questions.Select2Question()

    form.Form.set_resource_url("static")
    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "static/select2.min.js" in test_form.extra_js[0]
    form.Form.set_resource_url(SURVEY_JS_CDN)


def test_construct_survey_with_extra_css_and_set_resource_url():
    class TestForm(form.Form):
        text1 = questions.TagBoxQuestion()

    form.Form.set_resource_url("static")
    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert "text1" in test_form._form_elements
    assert "static/select2.min.css" in test_form.extra_css[0]
    form.Form.set_resource_url(SURVEY_JS_CDN)


def test_construct_survey_no_questions_elements():
    class TestForm(form.Form):
        text1 = "just text"

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert isinstance(survey, questions.Survey)
    assert test_form._form_elements == {}


def test_construct_survey_with_pages():
    class PageForm(form.Form):
        text1 = questions.TextQuestion()

    class TestForm(form.Form):
        page1 = form.FormPage(PageForm, name="Page1")

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert survey.pages[0].name == "Page1"
    assert survey.pages[0].questions[0].name == "text1"


def test_construct_survey_with_pages_no_page_name():
    class PageForm(form.Form):
        text1 = questions.TextQuestion()

    class TestForm(form.Form):
        page1 = form.FormPage(PageForm)

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert survey.pages[0].name == "PageForm"
    assert survey.pages[0].questions[0].name == "text1"


def test_construct_survey_with_second_level_panel():
    class PanelForm(form.Form):
        text1 = questions.TextQuestion()

    class PageForm(form.Form):
        panel1 = form.FormPanel(PanelForm, name="Panel1")

    class TestForm(form.Form):
        page1 = form.FormPage(PageForm)

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert survey.pages[0].name == "PageForm"
    assert survey.pages[0].questions[0].name == "Panel1"
    assert survey.pages[0].questions[0].elements[0].name == "text1"


def test_construct_survey_with_second_level_panel_no_name():
    class PanelForm(form.Form):
        text1 = questions.TextQuestion()

    class PageForm(form.Form):
        panel1 = form.FormPanel(PanelForm)

    class TestForm(form.Form):
        page1 = form.FormPage(PageForm)

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert survey.pages[0].name == "PageForm"
    assert survey.pages[0].questions[0].name == "PanelForm"
    assert survey.pages[0].questions[0].elements[0].name == "text1"


def test_construct_survey_with_second_level_panel_dynamic():
    class PanelForm(form.Form):
        text1 = questions.TextQuestion()

    class PageForm(form.Form):
        panel1 = form.FormPanel(PanelForm, name="Panel1", dynamic=True)

    class TestForm(form.Form):
        page1 = form.FormPage(PageForm)

    test_form = TestForm(name="testing")
    survey = test_form._construct_survey()
    assert survey.pages[0].name == "PageForm"
    assert survey.pages[0].questions[0].name == "Panel1"
    assert survey.pages[0].questions[0].template_elements[0].name == "text1"


def test_validate_no_validators():
    class TestForm(form.Form):
        text1 = questions.TextQuestion()

    test_form = TestForm(name="testing")
    form_data = {}
    assert test_form.validate(form_data) is True


def test_validate_required_valid():
    class TestForm(form.Form):
        text1 = questions.TextQuestion(required=True)

    test_form = TestForm(name="testing")
    form_data = {"text1": "hello"}
    assert test_form.validate(form_data) is True


def test_validate_required_invalid():
    class TestForm(form.Form):
        text1 = questions.TextQuestion(required=True)

    test_form = TestForm(name="testing")
    form_data = {}
    assert test_form.validate(form_data) is False


def test_validate_with_validators_valid():
    text_validator = TextValidator(min_length=5)

    class TestForm(form.Form):
        text1 = questions.TextQuestion(validators=[text_validator])

    test_form = TestForm(name="testing")
    form_data = {"text1": "you had me at hello"}
    assert test_form.validate(form_data) is True
    assert "__errors__" not in form_data


def test_validate_with_validators_invalid_with_set_errors():
    text_validator = TextValidator(min_length=5)

    class TestForm(form.Form):
        text1 = questions.TextQuestion(validators=[text_validator])

    test_form = TestForm(name="testing")
    form_data = {"text1": "Bye"}
    assert test_form.validate(form_data, set_errors=True) is False
    assert form_data["__errors__"][0]["question"] == "text1"


def test_update_object():
    text_validator = TextValidator(min_length=5)

    class AnyObj(object):
        def __init__(self, value):
            self.text1 = value

    class TestForm(form.Form):
        text1 = questions.TextQuestion(validators=[text_validator])

    test_form = TestForm(name="testing")
    form_data = {"text1": "hello"}
    anyobj = AnyObj("xxx")
    assert anyobj.text1 == "xxx"
    test_form.update_object(anyobj, form_data)
    assert anyobj.text1 == "hello"


def test_update_object_validation_error():
    text_validator = TextValidator(min_length=5)

    class AnyObj(object):
        def __init__(self, value):
            self.text1 = value

    class TestForm(form.Form):
        text1 = questions.TextQuestion(validators=[text_validator])

    test_form = TestForm(name="testing")
    form_data = {"text1": "bad"}
    anyobj = AnyObj("xxx")
    assert anyobj.text1 == "xxx"
    with pytest.raises(ValidationError):
        test_form.update_object(anyobj, form_data)
    assert anyobj.text1 == "xxx"



"""Defines constants to be used throughout the code."""

SURVEY_JS_VERSION = "1.9.81"

SURVEY_JS_CDN = "https://unpkg.com"

SURVEY_JS_WIDGETS = "surveyjs-widgets/surveyjs-widgets.min.js"

SURVEY_JS_PLATFORMS = (
    "angular",
    "jquery",
    "knockout",
    "react",
    "vue",
)

SUGGESTED_JS_BY_PLATFORM = {
    "angular": [
        "https://npmcdn.com/zone.js",
        "https://npmcdn.com/core-js@2.6.5/client/shim.min.js",
        "https://npmcdn.com/rxjs@5.0.0-beta.6/bundles/Rx.umd.js",
        "https://npmcdn.com/@angular/core@2.0.0-rc.5/bundles/core.umd.js",
        "https://npmcdn.com/@angular/common@2.0.0-rc.5/bundles/common.umd.js",
        "https://npmcdn.com/@angular/compiler@2.0.0-rc.5/bundles/compiler.umd.js",
        "https://npmcdn.com/@angular/platform-browser@2.0.0-rc.5/bundles/platform-browser.umd.js",
        "https://npmcdn.com/@angular/platform-browser-dynamic@2.0.0-rc.5/bundles/platform-browser-dynamic.umd.js",
    ],
    "jquery": [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
    ],
    "knockout": [
        "https://cdnjs.cloudflare.com/ajax/libs/knockout/3.4.0/knockout-min.js",
    ],
    "react": [
        "https://cdnjs.cloudflare.com/ajax/libs/babel-polyfill/7.6.0/polyfill.js",
        "https://unpkg.com/react@15/dist/react.js",
        "https://unpkg.com/react-dom@15/dist/react-dom.js",
        "https://unpkg.com/@babel/standalone@7.2.5/babel.min.js",
    ],
    "vue": [
        "https://unpkg.com/vue/dist/vue.js",
    ],
}

BOOTSTRAP_URL = "https://unpkg.com/bootstrap@3.3.7/dist/css/bootstrap.min.css"

INCLUDE_KEYS = {
    "pages": {
        "__all__": {
            "name": ...,
            "questions": {
                "__all__": {
                    "accepted_types": ...,
                    "add_row_location": ...,
                    "add_row_text": ...,
                    "allow_add_panel": ...,
                    "allow_add_rows": ...,
                    "allow_remove_panel": ...,
                    "allow_remove_rows": ...,
                    "allow_clear": ...,
                    "allow_images_preview": ...,
                    "allow_multiple": ...,
                    "all_rows_required": ...,
                    "auto_close": ...,
                    "auto_unmask": ...,
                    "bg_emotion": ...,
                    "cells": ...,
                    "cell_type": ...,
                    "choices": ...,
                    "choices_by_url": ...,
                    "choices_enable_if": ...,
                    "choices_max": ...,
                    "choices_min": ...,
                    "choices_order": ...,
                    "choices_step": ...,
                    "choices_visible_if": ...,
                    "clear_button": ...,
                    "columns": ...,
                    "columns_visible_if": ...,
                    "column_col_count": ...,
                    "column_layout": ...,
                    "column_min_width": ...,
                    "cols": ...,
                    "col_count": ...,
                    "config": ...,
                    "confirm_delete": ...,
                    "confirm_delete_text": ...,
                    "content_mode": ...,
                    "correct_answer": ...,
                    "currency": ...,
                    "date_format": ...,
                    "days_of_week_highlighted": ...,
                    "default_row_value": ...,
                    "default_value": ...,
                    "default_value_from_last_panel": ...,
                    "default_value_from_last_row": ...,
                    "description": ...,
                    "description_location": ...,
                    "direction": ...,
                    "disable_touch_keyboard": ...,
                    "display_style": ...,
                    "elements": ...,
                    "emotion_color": ...,
                    "emotion_size": ...,
                    "emotions": ...,
                    "emotions_count": ...,
                    "empty_text": ...,
                    "enable_if": ...,
                    "end_date": ...,
                    "expression": ...,
                    "expression_format": ...,
                    "has_none": ...,
                    "has_other": ...,
                    "has_select_all": ...,
                    "height": ...,
                    "hide_if_choices_empty": ...,
                    "hide_number": ...,
                    "horizontal_scroll": ...,
                    "html": ...,
                    "image_fit": ...,
                    "image_height": ...,
                    "image_width": ...,
                    "indent": ...,
                    "inner_indent": ...,
                    "input_format": ...,
                    "input_mask": ...,
                    "input_type": ...,
                    "items": ...,
                    "item_size": ...,
                    "key_duplication_error": ...,
                    "key_name": ...,
                    "kind": ...,
                    "label_false": ...,
                    "label_true": ...,
                    "maximum_fraction_digits": ...,
                    "max_answers_count": ...,
                    "max_date": ...,
                    "max_length": ...,
                    "max_panel_count": ...,
                    "max_rate_description": ...,
                    "max_row_count": ...,
                    "max_size": ...,
                    "max_value": ...,
                    "max_width": ...,
                    "minimum_fraction_digits": ...,
                    "min_date": ...,
                    "min_panel_count": ...,
                    "min_rate_description": ...,
                    "min_row_count": ...,
                    "min_width": ...,
                    "min_value": ...,
                    "multi_select": ...,
                    "name": ...,
                    "need_confirm_remove_file": ...,
                    "none_text": ...,
                    "orientation": ...,
                    "options_caption": ...,
                    "other_text": ...,
                    "other_error_text": ...,
                    "other_place_holder": ...,
                    "panel_add_text": ...,
                    "panel_count": ...,
                    "panel_next_text": ...,
                    "panel_prev_text": ...,
                    "panel_remove_text": ...,
                    "panels_state": ...,
                    "pips_mode": ...,
                    "pips_density": ...,
                    "pips_text": ...,
                    "pips_values": ...,
                    "place_holder": ...,
                    "prefix": ...,
                    "question_start_index": ...,
                    "question_title_location": ...,
                    "range_max": ...,
                    "range_min": ...,
                    "rate_max": ...,
                    "rate_min": ...,
                    "rate_step": ...,
                    "rating_theme": ...,
                    "remove_row_text": ...,
                    "render_as": ...,
                    "render_mode": ...,
                    "required": ...,
                    "required_error_text": ...,
                    "required_if": ...,
                    "rows": ...,
                    "rows_order": ...,
                    "rows_visible_if": ...,
                    "row_count": ...,
                    "row_title_width": ...,
                    "select_all_text": ...,
                    "select2_config": ...,
                    "show_clear_button": ...,
                    "show_header": ...,
                    "show_label": ...,
                    "show_number": ...,
                    "show_options_caption": ...,
                    "show_preview": ...,
                    "show_question_numbers": ...,
                    "show_range_in_progress": ...,
                    "show_title": ...,
                    "show_values": ...,
                    "size": ...,
                    "start_date": ...,
                    "start_with_new_line": ...,
                    "state": ...,
                    "step": ...,
                    "store_data_as_text": ...,
                    "template_elements": ...,
                    "template_description": ...,
                    "template_title": ...,
                    "template_title_location": ...,
                    "text_update_mode": ...,
                    "title": ...,
                    "title_location": ...,
                    "today_highlight": ...,
                    "tooltips": ...,
                    "total_text": ...,
                    "use_display_values_in_title": ...,
                    "use_grouping": ...,
                    "validators": ...,
                    "value_false": ...,
                    "value_name": ...,
                    "value_true": ...,
                    "visible": ...,
                    "visible_if": ...,
                    "wait_for_upload": ...,
                    "week_start": ...,
                    "width": ...,
                }
            },
            "title": ...,
            "max_time_to_finish": ...,
            "navigation_buttons_visibility": ...,
            "question_title_location": ...,
            "questions_order": ...,
        }
    },
    "title": ...,
    "calculated_values": ...,
    "check_errors_mode": ...,
    "clear_invisible_values": ...,
    "completed_before_html": ...,
    "completed_html": ...,
    "completed_html_on_condition": ...,
    "complete_text": ...,
    "cookie_name": ...,
    "description": ...,
    "edit_text": ...,
    "first_page_is_started": ...,
    "focus_first_question_automatic": ...,
    "focus_on_first_error": ...,
    "go_next_page_automatic": ...,
    "loading_html": ...,
    "locale": ...,
    "logo": ...,
    "logo_fit": ...,
    "logo_height": ...,
    "logo_position": ...,
    "logo_width": ...,
    "max_others_length": ...,
    "max_text_length": ...,
    "max_time_to_finish": ...,
    "max_time_to_finish_page": ...,
    "mode": ...,
    "navigate_to_url": ...,
    "navigate_to_url_on_condition": ...,
    "page_next_text": ...,
    "page_prev_text": ...,
    "preview_text": ...,
    "progress_bar_type": ...,
    "questions_on_page_mode": ...,
    "questions_order": ...,
    "question_description_location": ...,
    "question_error_location": ...,
    "question_start_index": ...,
    "question_title_location": ...,
    "question_title_pattern": ...,
    "question_title_template": ...,
    "required_text": ...,
    "send_result_on_page_next": ...,
    "show_completed_page": ...,
    "show_navigation_buttons": ...,
    "show_page_numbers": ...,
    "show_page_titles": ...,
    "show_prev_button": ...,
    "show_preview_before_complete": ...,
    "show_progress_bar": ...,
    "show_question_numbers": ...,
    "show_timer_panel": ...,
    "show_timer_panel_mode": ...,
    "show_title": ...,
    "start_survey_text": ...,
    "store_others_as_comment": ...,
    "survey_id": ...,
    "survey_post_id": ...,
    "survey_show_data_saving": ...,
    "text_update_mode": ...,
    "triggers": ...,
}

TEXT_INPUT_TYPES = (
    "color",
    "date",
    "datetime",
    "datetime-local",
    "email",
    "month",
    "number",
    "password",
    "range",
    "tel",
    "text",
    "time",
    "url",
    "week",
)

LOCALES = (
    "ar",
    "bg",
    "ca",
    "cs",
    "da",
    "de",
    "en",
    "es",
    "et",
    "fa",
    "fi",
    "fr",
    "gr",
    "he",
    "hu",
    "id",
    "is",
    "it",
    "ja",
    "ka",
    "ko",
    "lt",
    "lv",
    "nl",
    "no",
    "pl",
    "pt",
    "ro",
    "ru",
    "sv",
    "sw",
    "tg",
    "th",
    "tr",
    "ua",
    "zh-cn",
    "zh-tw",
)

CHOICE_ORDER_VALUES = (
    "none",
    "asc",
    "desc",
    "random",
)

CHOICES_BY_URL_KEYS = (
    "path",
    "titleName",
    "url",
    "valueName",
)

IMAGE_FIT_VALUES = (
    "contain",
    "none",
    "cover",
    "fill",
)

IMAGE_CONTENT_MODE_VALUES = (
    "image",
    "video",
)

PANEL_RENDER_MODES = (
    "progressTop",
    "progressBottom",
)

DESCRIPTION_LOCATIONS = (
    "default",
    "underInput",
    "underTitle",
)

TITLE_LOCATIONS = (
    "default",
    "top",
    "bottom",
    "left",
    "hidden",
)

CHECK_ERRORS_MODES = (
    "onNextPage",
    "onValueCHanged",
    "onComplete",
)

TEXT_UPDATE_MODES = (
    "default",
    "onBlur",
    "onTyping",
)

EXPRESSION_DISPLAY_STYLES = (
    "none",
    "decimal",
    "currency",
    "percent",
    "date",
)

SHOW_QUESTION_NUMBERS_VALUES = (
    "default",
    "onpanel",
    "off",
)

PANEL_STATES = (
    "default",
    "collapsed",
    "expanded",
    "firstExpanded",
)

QUESTION_ORDER_VALUES = (
    "default",
    "initial",
    "random",
)

NAV_BUTTONS_VISIBILITY = (
    "inherit",
    "show",
    "hide",
)

NAV_BUTTONS_POSITIONS = (
    "none",
    "top",
    "bottom",
    "both",
)

CLEAR_INVISIBLE_VALUES = (
    "none",
    "onHidden",
    "onComplete",
)

LOGO_POSITIONS = (
    "none",
    "left",
    "right",
    "top",
    "bottom",
)

PROGRESS_BAR_TYPES = (
    "pages",
    "questions",
    "correctQuestions",
)

SURVEY_MODES = (
    "edit",
    "display",
)

QUESTION_DESCRIPTION_LOCATIONS = (
    "underTitle",
    "underInput",
)

QUESTION_ERROR_LOCATIONS = (
    "top",
    "bottom",
)

QUESTION_PAGE_MODES = (
    "standard",
    "singlePage",
    "questionPerPage",
)

PAGE_SHOW_QUESTION_NUMBERS_VALUES = (
    "on",
    "onpage",
    "off",
)

SHOW_PROGRESS_BAR_OPTIONS = (
    "off",
    "top",
    "bottom",
    "both",
)

SHOW_PREVIEW_VALUES = (
    "noPreview",
    "showAllQuestions",
    "showAnsweredQuestions",
)

SHOW_TIMER_VALUES = (
    "none",
    "top",
    "bottom",
)

SHOW_TIMER_MODES = (
    "page",
    "survey",
    "all",
)

SURVEY_TEXT_UPDATE_MODES = (
    "onBlur",
    "onTyping",
)

ROW_ORDER_VALUES = (
    "initial",
    "random",
)

MATRIX_CELL_TYPES = (
    "dropdown",
    "checkbox",
    "radiogroup",
    "text",
    "comment",
    "boolean",
    "expression",
    "rating",
)

MATRIX_COLUMN_LAYOUTS = (
    "horizontal",
    "vertical",
)

MATRIX_ROW_LOCATIONS = (
    "default",
    "top",
    "bottom",
    "topBottom",
)

SURVEY_JS_THEMES = (
    "defaultV2",
    "bootstrap",
    "modern",
)

BAR_RATING_THEMES = (
    "fontawesome-stars",
    "css-stars",
    "bars-pill",
    "bars-1to10",
    "bars-movie",
    "bars-reversed",
    "bars-horizontal",
    "fontawesome-stars-o",
)
